import csv
import logging
import sys
import tempfile
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Tuple

import csp
import orjson
import pyarrow as pa
import pyarrow.parquet as pq
import pytest

if not sys.platform.startswith("linux"):
    pytest.skip("Skipping Linux-specific tests on macOS platforms", allow_module_level=True)

from csp import ts

from csp_gateway import AddChannelsToGraphOutput, FileDropType, GatewayChannels, GatewayModule, GatewayStruct, ReadFileDrop, ReadFileDropConfiguration
from csp_gateway.testing.shared_helpful_classes import (
    MyGateway,
)


class FDStruct(GatewayStruct):
    i: int
    s: str
    b: bool
    f: float
    l_i: List[int]


class NotFDStruct(GatewayStruct):
    i: str
    s: int


class FDGatewayChannels(GatewayChannels):
    fd_channel: ts[FDStruct] = None
    fd_list_channel: ts[List[FDStruct]] = None
    fd_list_channel_2: ts[List[FDStruct]] = None
    fd_dict_channel: ts[Dict[str, FDStruct]] = None
    fd_dict_basket_channel: Dict[str, ts[FDStruct]] = None

    def dynamic_keys(self):
        return {FDGatewayChannels.fd_dict_basket_channel: ["a", "b", "c"]}


def convert_to_dict(data):
    new_data = []
    for d in data:
        if isinstance(d, list):
            new_data.append(convert_to_dict(d))
        elif isinstance(d, dict):
            new_data.append({k: v.to_dict() for k, v in d.items()})
        else:
            new_data.append(d.to_dict())
    return new_data


def csv_writer(path, data):
    dict_data = convert_to_dict(data)
    fieldnames = []
    for d in dict_data:
        fieldnames.extend(list(d.keys()))
    fieldnames = list(set(fieldnames))
    with open(path, "w") as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(dict_data)


def json_writer(path, data, write_bad_file=False):
    dict_data = convert_to_dict(data)
    with open(path, "wb") as f:
        data_bytes = orjson.dumps(dict_data)
        if write_bad_file:
            data_bytes = data_bytes[: -1 * int(len(data_bytes) / 3)]
        f.write(data_bytes)


def json_writer_bad(path, data):
    json_writer(path, data, True)


def parquet_writer(path, data):
    table = pa.Table.from_pylist(convert_to_dict(data))
    pq.write_table(table, path)


def get_all_writers():
    return [(csv_writer, FileDropType.CSV), (json_writer, FileDropType.JSON), (parquet_writer, FileDropType.PARQUET)]


class Writer(GatewayModule):
    data: List[Tuple[float, object, str, object]]

    def connect(self, channels):
        self.execute()

    @csp.node
    def execute(self):
        with csp.alarms():
            alarm = csp.alarm(int)

        with csp.start():
            csp.schedule_alarm(alarm, timedelta(seconds=self.data[0][0]), 0)

        if csp.ticked(alarm):
            cur_idx = alarm
            next_idx = cur_idx + 1
            if next_idx < len(self.data):
                csp.schedule_alarm(alarm, timedelta(seconds=self.data[next_idx][0]), next_idx)
            _, write_fn, filename, structs = self.data[cur_idx]
            write_fn(filename, structs)


def match_data(data1, data2, exact=False):
    assert len(data1) == len(data2)
    for d1, d2 in zip(data1, data2):
        dict1 = d1.to_dict()
        dict2 = d2.to_dict()
        if not exact:
            dict1.pop("id")
            dict1.pop("timestamp")
            dict2.pop("id")
            dict2.pop("timestamp")
        assert dict1 == dict2


def match_lists(lists_data, data2, exact=False):
    for data1 in lists_data:
        match_data(data1, data2, exact)


@pytest.mark.parametrize(
    "structs",
    [
        [FDStruct(i=i) for i in range(1)],
        [FDStruct(i=i) for i in range(10)],
        [FDStruct(i=i) for i in range(100)],
    ],
)
def test_basic(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=FileDropType.JSON),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs)


@pytest.mark.parametrize(
    "structs",
    [
        [[FDStruct(i=i) for i in range(1)]],
        [[FDStruct(i=i) for i in range(10)]],
        [[FDStruct(i=i) for i in range(100)]],
    ],
)
def test_list_of_structs(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        out_data = [d[1] for d in out["fd_list_channel"]]
        assert len(out_data) == 1
        match_lists(out_data, structs[0])


@pytest.mark.parametrize("structs", [[[FDStruct(i=i) for i in range(10)]]])
@pytest.mark.parametrize("no_readers", [1, 2, 5])
def test_multi_readers_single_channel_single_dir(structs, no_readers):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
            ]
            * no_readers,
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        out_data = [d[1] for d in out["fd_list_channel"]]
        assert len(out_data) == no_readers
        match_lists(out_data, structs[0])


@pytest.mark.parametrize("structs", [[[FDStruct(i=i) for i in range(10)]]])
def test_multi_readers_multi_channel_single_directory(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel_2", filedrop_type=FileDropType.JSON),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        out_data_1 = [d[1] for d in out["fd_list_channel"]]
        out_data_2 = [d[1] for d in out["fd_list_channel_2"]]
        assert len(out_data_1) == 1
        assert len(out_data_2) == 1
        match_lists(out_data_1, structs[0])
        match_lists(out_data_2, structs[0])


@pytest.mark.parametrize(
    "structs",
    [
        [[FDStruct(i=i) for i in range(1)]],
        [[FDStruct(i=i) for i in range(10)]],
        [[FDStruct(i=i) for i in range(100)]],
    ],
)
def test_single_channel_multi_dir(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir1:
        with tempfile.TemporaryDirectory(dir=".") as dir2:
            dirpath1 = Path(dir1)
            dirpath2 = Path(dir2)
            writer = Writer(
                data=[[1, json_writer, str(dirpath1 / "json_file1.json"), structs], [0, json_writer, str(dirpath2 / "json_file2.json"), structs]]
            )
            fd_module = ReadFileDrop(
                configs=[
                    ReadFileDropConfiguration(dir_path=dirpath1, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
                    ReadFileDropConfiguration(dir_path=dirpath2, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
                ]
            )
            gateway = MyGateway(
                modules=[writer, fd_module, AddChannelsToGraphOutput()],
                channels=FDGatewayChannels(),
            )
            out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
            out_data = [d[1] for d in out["fd_list_channel"]]
            assert len(out_data) == 2
            match_lists(out_data, structs[0])


@pytest.mark.parametrize(
    "structs",
    [
        [[FDStruct(i=i) for i in range(1)]],
        [[FDStruct(i=i) for i in range(10)]],
        [[FDStruct(i=i) for i in range(100)]],
    ],
)
def test_multi_channel_multi_dir(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir1:
        with tempfile.TemporaryDirectory(dir=".") as dir2:
            dirpath1 = Path(dir1)
            dirpath2 = Path(dir2)
            writer = Writer(
                data=[[1, json_writer, str(dirpath1 / "json_file1.json"), structs], [0, json_writer, str(dirpath2 / "json_file2.json"), structs]]
            )
            fd_module = ReadFileDrop(
                configs=[
                    ReadFileDropConfiguration(dir_path=dirpath1, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON),
                    ReadFileDropConfiguration(dir_path=dirpath2, channel_name="fd_list_channel_2", filedrop_type=FileDropType.JSON),
                ]
            )
            gateway = MyGateway(
                modules=[writer, fd_module, AddChannelsToGraphOutput()],
                channels=FDGatewayChannels(),
            )
            out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
            out_data_1 = [d[1] for d in out["fd_list_channel"]]
            out_data_2 = [d[1] for d in out["fd_list_channel_2"]]
            assert len(out_data_1) == 1
            assert len(out_data_2) == 1
            match_lists(out_data_1, structs[0])
            match_lists(out_data_2, structs[0])


@pytest.mark.parametrize(
    "structs",
    [
        [FDStruct(i=i) for i in range(1)],
        [FDStruct(i=i) for i in range(10)],
    ],
)
@pytest.mark.parametrize(
    "filetype_data",
    [
        [json_writer, "file1.json", FileDropType.JSON],
        [parquet_writer, "file1.parquet", FileDropType.PARQUET],
    ],
)
def test_filetypes(structs, filetype_data):
    writer = filetype_data[0]
    filename = filetype_data[1]
    filedrop_type = filetype_data[2]
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, writer, str(dirpath / filename), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=filedrop_type),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs)


def test_invalid_data(caplog):
    structs = [FDStruct(i=i) for i in range(1)]
    with caplog.at_level(logging.ERROR):
        with tempfile.TemporaryDirectory(dir=".") as dir:
            dirpath = Path(dir)
            writer = Writer(data=[[1, json_writer_bad, str(dirpath / "file.json"), structs]])
            fd_module = ReadFileDrop(
                configs=[
                    ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=FileDropType.JSON),
                ]
            )
            gateway = MyGateway(
                modules=[writer, fd_module, AddChannelsToGraphOutput()],
                channels=FDGatewayChannels(),
            )
            _ = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
    assert len(caplog.text) > 0
    assert "Failed to read data" in caplog.text


def test_invalid_struct(caplog):
    structs = [NotFDStruct(i="a") for i in range(1)]
    with caplog.at_level(logging.ERROR):
        with tempfile.TemporaryDirectory(dir=".") as dir:
            dirpath = Path(dir)
            writer = Writer(data=[[1, json_writer, str(dirpath / "file.json"), structs]])
            fd_module = ReadFileDrop(
                configs=[
                    ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=FileDropType.JSON),
                ]
            )
            gateway = MyGateway(
                modules=[writer, fd_module, AddChannelsToGraphOutput()],
                channels=FDGatewayChannels(),
            )
            _ = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
    assert len(caplog.text) > 0
    assert "Failed to read data" in caplog.text


@pytest.mark.parametrize("writer_data", get_all_writers())
@pytest.mark.parametrize(
    "filedata",
    [
        [[0.1, f"file{idx}", [FDStruct(i=val) for val in range(idx + 1)]] for idx in range(2)],
        [[0.01, f"file{idx}", [FDStruct(i=val) for val in range(idx + 1)]] for idx in range(10)],
        [[0.01, f"file{idx}", [FDStruct(i=val) for val in range(idx + 1)]] for idx in range(100)],
    ],
)
def test_multiple_files(writer_data, filedata):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[fdata[0], writer_data[0], str(dirpath / fdata[1]), fdata[2]] for fdata in filedata])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=writer_data[1]),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        s = sum([len(fd[2]) for fd in filedata])
        assert len(out["fd_channel"]) == s
        csp_data = [d[1] for d in out["fd_channel"]]
        fd = []
        for fd_data in filedata:
            fd.extend(fd_data[2])
        match_data(csp_data, fd)


@pytest.mark.parametrize(
    "structs",
    [
        [FDStruct(i=i, s=chr(ord("a") + i % 26)) for i in range(1)],
        [FDStruct(i=i, s=chr(ord("a") + i % 26)) for i in range(10)],
        [FDStruct(i=i, s=chr(ord("a") + i % 26)) for i in range(100)],
    ],
)
def test_config_options(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(
                    dir_path=dirpath,
                    channel_name="fd_channel",
                    filedrop_type=FileDropType.JSON,
                    subscribe_with_struct_timestamp=True,
                    subscribe_with_struct_id=True,
                ),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs, exact=True)


def test_extensions():
    structs_1 = [[FDStruct(i=i) for i in range(10)]]
    structs_2 = [[FDStruct(i=i) for i in range(20)]]
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(
            data=[[0.1, json_writer, str(dirpath / "json_file1.js1"), structs_1], [1, json_writer, str(dirpath / "json_file1.js2"), structs_2]]
        )
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel", filedrop_type=FileDropType.JSON, extensions=[".js1"]),
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_list_channel_2", filedrop_type=FileDropType.JSON, extensions=[".js2"]),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        out_data_1 = [d[1] for d in out["fd_list_channel"]]
        out_data_2 = [d[1] for d in out["fd_list_channel_2"]]
        assert len(out_data_1) == 1
        assert len(out_data_2) == 1
        match_lists(out_data_1, structs_1[0])
        match_lists(out_data_2, structs_2[0])


@pytest.mark.parametrize(
    "structs",
    [
        [{"a": FDStruct(i=i)} for i in range(1)],
        [{"a": FDStruct(i=i)} for i in range(10)],
        [{"a": FDStruct(i=i)} for i in range(100)],
    ],
)
def test_dict(structs):
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_dict_channel", filedrop_type=FileDropType.JSON),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        csp_data = [d[1] for d in out["fd_dict_channel"]]
        for d in csp_data:
            assert "a" in d
        csp_data = [d["a"] for d in csp_data]
        data = [d["a"] for d in structs]
        match_data(csp_data, data)

    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_dict_basket_channel", filedrop_type=FileDropType.JSON),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        csp_data = [d[1] for d in out["fd_dict_basket_channel[a]"]]
        data = [d["a"] for d in structs]
        match_data(csp_data, data)


@pytest.mark.parametrize(
    "structs",
    [
        [FDStruct(i=i) for i in range(1)],
        [FDStruct(i=i) for i in range(10)],
    ],
)
def test_filedrop_type_custom(structs):
    def custom_writer(path, data):
        dict_data = convert_to_dict(data)
        new_dict_data = []
        for d in dict_data:
            new_d = {}
            for k, v in d.items():
                new_d[k.upper()] = v
            new_dict_data.append(new_d)
        with open(path, "wb") as f:
            data_bytes = orjson.dumps(new_dict_data)
            f.write(data_bytes)

    def custom_loader(path):
        with open(path, "rb") as f:
            dict_data = orjson.loads(f.read())
        dict_data = [{k.lower(): v for k, v in d.items()} for d in dict_data]
        return dict_data

    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, custom_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=FileDropType.CUSTOM, loader=custom_loader),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs)


@pytest.mark.parametrize(
    "structs",
    [
        [FDStruct(i=i) for i in range(1)],
        [FDStruct(i=i) for i in range(10)],
    ],
)
def test_deserializer(structs):
    def custom_writer(path, data):
        dict_data = convert_to_dict(data)
        new_dict_data = []
        for d in dict_data:
            new_d = {}
            for k, v in d.items():
                new_d[k.upper()] = v
            new_dict_data.append(new_d)
        with open(path, "wb") as f:
            data_bytes = orjson.dumps(new_dict_data)
            f.write(data_bytes)

    def deserializer(data):
        new_data = {k.lower(): v for k, v in data.items()}
        new_data.pop("id")
        new_data.pop("timestamp")
        return FDStruct(**new_data)

    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, custom_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type=FileDropType.JSON, deserializer=deserializer),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs)


def test_fieldrop_type_as_str():
    structs = [FDStruct(i=i) for i in range(1)]
    with tempfile.TemporaryDirectory(dir=".") as dir:
        dirpath = Path(dir)
        writer = Writer(data=[[1, json_writer, str(dirpath / "json_file1.json"), structs]])
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filedrop_type="JSON"),
            ]
        )
        gateway = MyGateway(
            modules=[writer, fd_module, AddChannelsToGraphOutput()],
            channels=FDGatewayChannels(),
        )
        out = csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))
        match_data([d[1] for d in out["fd_channel"]], structs)

    with pytest.raises(ValueError):
        ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filetype_data="NOT_A_TYPE")

    with pytest.raises(ValueError):
        ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filetype_data="CUSTOM")
    with pytest.raises(ValueError):
        ReadFileDropConfiguration(dir_path=dirpath, channel_name="fd_channel", filetype_data=FileDropType.CUSTOM, loader=None)


def test_invalid_channels():
    class MyFDGatewayChannels(FDGatewayChannels):
        my_bad_dict_channel: Dict[str, str] = None
        my_bad_list_channel: List[str] = None
        my_bad_simple_channel: str = None

    def dynamic_keys(self):
        return {MyFDGatewayChannels.my_bad_dict_channel: ["a", "b", "c"]}

    with tempfile.TemporaryDirectory(dir=".") as dir:
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=Path(dir), channel_name="my_bad_dict_channel", filedrop_type="JSON"),
            ]
        )
        gateway = MyGateway(
            modules=[fd_module, AddChannelsToGraphOutput()],
            channels=MyFDGatewayChannels(),
        )
        with pytest.raises(ValueError, match=r".*should be of the form Dict\[KeyType, TsType\].*"):
            csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))

    with tempfile.TemporaryDirectory(dir=".") as dir:
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=Path(dir), channel_name="my_bad_list_channel", filedrop_type="JSON"),
            ]
        )
        gateway = MyGateway(
            modules=[fd_module, AddChannelsToGraphOutput()],
            channels=MyFDGatewayChannels(),
        )
        with pytest.raises(ValueError, match=r".*should be of the form List\[TsType\].*"):
            csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))

    with tempfile.TemporaryDirectory(dir=".") as dir:
        fd_module = ReadFileDrop(
            configs=[
                ReadFileDropConfiguration(dir_path=Path(dir), channel_name="my_bad_simple_channel", filedrop_type="JSON"),
            ]
        )
        gateway = MyGateway(
            modules=[fd_module, AddChannelsToGraphOutput()],
            channels=MyFDGatewayChannels(),
        )
        with pytest.raises(Exception, match=r".*Channel type cannot be handled.*"):
            csp.run(gateway.graph, realtime=True, endtime=timedelta(seconds=5))



from datetime import datetime, timezone
from types import MappingProxyType
from typing import Any, Dict

import csp
from csp import Struct
from pydantic import ValidationInfo
from pydantic_core import CoreConfig, core_schema

from ..id_generator import get_counter
from .psp import PerspectiveUtilityMixin

IdType = str

__all__ = (
    "GatewayStruct",
    "IdType",
    "GatewayLookupMixin",
    "GatewayPydanticMixin",
    "GatewayStructMixins",
    "is_gateway_struct_like",
)


class GatewayLookupMixin:
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.id_generator = get_counter(cls)
        cls._internal_mapping: Dict[str, Any] = {}
        cls.lookup = MappingProxyType(cls._internal_mapping).get
        cls._include_in_lookup = True

    def __init__(self, **kwargs: Any) -> None:
        if "id" not in kwargs:
            kwargs["id"] = str(self.__class__.id_generator.next())
        if "timestamp" not in kwargs:
            kwargs["timestamp"] = datetime.now(timezone.utc)
        if getattr(self.__class__, "_include_in_lookup", True):
            # Insert into lookup before super to keep behavior consistent
            # with previous GatewayStruct construction
            # (instance becomes available immediately after init)
            # Self will be fully initialized once super returns
            self.__class__._internal_mapping[kwargs["id"]] = self
        super().__init__(**kwargs)

    @classmethod
    def omit_from_lookup(cls, omit=True):
        cls._include_in_lookup = not omit

    @classmethod
    def included_in_lookup(cls):
        return cls._include_in_lookup

    @classmethod
    def generate_id(cls) -> str:
        return str(cls.id_generator.next())


class GatewayPydanticMixin:
    @classmethod
    def _validate_gateway_struct_after(cls, val):
        """Validate GatewayStruct after pydantic type validation.
        A validator attached to every GatewayStruct to allow for defining custom
        model-level after validators that run after pydantic type validation.
        If not defined on a child class, the parent's validator will be used.  If defined on a child class, the parent's validator will be ignored. Please call the parent's validator directly if you want to run both.

        This is meant to be mixed-in with csp.Struct's. We do not inherit from a csp.Struct
        since csp.Struct's do not support multiple inheritance with other csp.Struct's

        Args:
            cls: The class this validator is attached to
            val: The value to validate
        Returns:
            The validated value, possibly modified
        """
        return val

    @classmethod
    def type_adapter(cls):
        # NOTE: Only needed until csp>0.9 is released with this fix
        # We mangle ourselves, explicitly, to make sure that child Structs
        # will get their own type adapters.
        attr_name = f"_{cls.__name__}__pydantic_type_adapter"
        internal_type_adapter = getattr(cls, attr_name, None)
        if internal_type_adapter:
            return internal_type_adapter

        # Late import to avoid autogen issues
        from pydantic import TypeAdapter

        type_adapter = TypeAdapter(cls)
        setattr(cls, attr_name, type_adapter)
        return type_adapter

    @classmethod
    def _validate_gateway_struct(cls, val, handler, info: ValidationInfo):
        if isinstance(info.context, dict) and isinstance(val, dict):
            if info.context.get("force_new_id", False):
                # If we are forcing a new id, we need to remove the old one
                val.pop("id", None)
            if info.context.get("force_new_timestamp", False):
                # If we are forcing a new timestamp, we need to remove the old one
                val.pop("timestamp", None)
        csp_struct = handler(val)
        final = cls._validate_gateway_struct_after(csp_struct)
        return final

    @staticmethod
    def _get_pydantic_core_schema(cls, source_type, handler):
        # Get parent schema - note the cls parameter
        parent_schema = csp.Struct._get_pydantic_core_schema(cls, source_type, handler)
        core_config = CoreConfig(coerce_numbers_to_str=True)
        # soooo hacky...
        parent_schema["schema"]["config"] = core_config
        return core_schema.with_info_wrap_validator_function(
            function=cls._validate_gateway_struct, schema=parent_schema, serialization=parent_schema.get("serialization")
        )


GatewayStructMixins = (GatewayLookupMixin, GatewayPydanticMixin, PerspectiveUtilityMixin)


class GatewayStruct(
    *GatewayStructMixins,
    Struct,
):
    """Convenience class composing gateway mixins with csp.Struct.

    Provides id/timestamp fields, lookup/registry utilities, and pydantic
    integration, plus Perspective utilities.
    """

    id: IdType
    timestamp: datetime


def is_gateway_struct_like(cls) -> bool:
    """Strict check: requires all gateway mixins and `csp.Struct`.

    Returns True only if `cls` is a `csp.Struct` subclass AND also
    subclasses `GatewayLookupMixin`, `GatewayPydanticMixin`, and
    `PerspectiveUtilityMixin`.
    """
    if not isinstance(cls, type):
        return False
    # Shortcut for explicit GatewayStruct
    if issubclass(cls, GatewayStruct):
        return True
    try:
        return (
            issubclass(cls, Struct)
            and issubclass(cls, GatewayLookupMixin)
            and issubclass(cls, GatewayPydanticMixin)
            and issubclass(cls, PerspectiveUtilityMixin)
        )
    except Exception:
        return False



import logging
import multiprocessing
import time
from datetime import datetime, timedelta
from io import StringIO
from typing import Any, Callable, Dict, List, Optional, Set, Type, Union

import csp
import numpy as np
import pytest
from ccflow import ModelRegistry
from csp import Enum, ts
from csp.typing import Numpy1DArray
from omegaconf import OmegaConf
from pydantic import Field, ValidationError

from csp_gateway import (
    Channels,
    ChannelSelection,
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewayStruct,
    State,
)
from csp_gateway.testing import GatewayTestHarness
from csp_gateway.utils import NoProviderException


class MyEnum(Enum):
    ONE = 1
    TWO = 2


class MyStruct(GatewayStruct):
    foo: float


class MyChildStruct(MyStruct):
    foo2: int


class MyGatewayChannels(GatewayChannels):
    my_channel_final: ts[MyStruct] = None
    my_channel_mid: ts[MyStruct] = None
    my_channel_dup: ts[MyStruct] = None

    my_keys: List[str] = []
    my_static: float = 0.0
    my_static_dict: Dict[str, float] = {}
    my_static_list: List[str] = []
    my_static_dict_of_objects: Dict[str, Any] = {}
    my_channel: ts[MyStruct] = None
    s_my_channel: ts[State[MyStruct]] = None
    my_list_channel: ts[List[MyStruct]] = None
    s_my_list_channel: ts[State[MyStruct]] = None
    my_enum_basket: Dict[MyEnum, ts[MyStruct]] = None
    my_str_basket: Dict[str, ts[MyStruct]] = None
    my_enum_basket_list: Dict[MyEnum, ts[List[MyStruct]]] = None
    my_array_channel: ts[Numpy1DArray[float]] = None

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: self.my_keys}


class MyGateway(Gateway):
    channels_model: Type[Channels] = MyGatewayChannels  # type: ignore[assignment]


class MySetModule(GatewayModule):
    my_data: ts[MyStruct]
    my_data2: ts[MyStruct]
    my_list_data: ts[List[MyStruct]]
    by_key: bool = True

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: ["my_key", "my_key2"]}

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        channels.set_channel(MyGatewayChannels.my_channel, self.my_data)
        channels.set_channel(
            MyGatewayChannels.my_list_channel,
            self.my_list_data,
        )
        channels.add_send_channel(MyGatewayChannels.my_channel)
        channels.set_state(MyGatewayChannels.my_channel, "id")
        channels.add_send_channel(MyGatewayChannels.my_list_channel)
        channels.set_state(MyGatewayChannels.my_list_channel, "id")

        channels.set_channel(MyGatewayChannels.my_array_channel, csp.const(np.array([1.0, 2.0])))

        if self.by_key:
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data, MyEnum.ONE)
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data2, MyEnum.TWO)
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data, "my_key")
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data2, "my_key2")
            channels.set_channel(MyGatewayChannels.my_enum_basket_list, self.my_list_data, MyEnum.ONE)
            channels.set_channel(MyGatewayChannels.my_enum_basket_list, self.my_list_data, MyEnum.TWO)
        else:
            channels.set_channel(
                MyGatewayChannels.my_enum_basket,
                {
                    MyEnum.ONE: self.my_data,
                    MyEnum.TWO: self.my_data2,
                },
            )
            channels.set_channel(
                MyGatewayChannels.my_str_basket,
                {
                    "my_key": self.my_data,
                    "my_key2": self.my_data2,
                },
            )
            channels.set_channel(
                MyGatewayChannels.my_enum_basket_list,
                {
                    MyEnum.ONE: self.my_list_data,
                    MyEnum.TWO: self.my_list_data,
                },
            )


class MyStatefulGetSetModule(GatewayModule):
    @csp.node
    def state_node_pre(
        self,
        data_to_send: ts[MyStruct],
    ) -> ts[MyStruct]:
        if csp.ticked(data_to_send):
            # Mimics doing some calculations
            return data_to_send.copy()

    def connect(self, channels: MyGatewayChannels) -> None:
        channels.add_send_channel(MyGatewayChannels.my_channel)
        res_pre = self.state_node_pre(
            channels.get_channel(MyGatewayChannels.my_channel),
        )
        # channels.set_channel(MyGatewayChannels.my_channel_mid, res_pre)
        res = self.state_node(
            channels.get_channel(MyGatewayChannels.my_channel_dup),
            res_pre,
        )
        channels.set_channel(MyGatewayChannels.my_channel_final, res)
        """
        This line below used to be problematic. If commented out, everything
        worked normally. If it was there, then, depending on the ordering
        of this module and the GatewayTestHarness, the ticks generated would
        sometimes occur in the same engine cycle, and sometimes not. This has been fixed.
        """
        channels.set_channel(MyGatewayChannels.my_channel_mid, res_pre)

    @csp.node
    def state_node(
        self,
        data_to_update: ts[MyStruct],
        data_to_send: ts[MyStruct],
    ) -> csp.Outputs(
        ts[MyStruct],
    ):
        val = []
        if csp.ticked(data_to_send):
            val.append(data_to_send.copy())

        if csp.ticked(data_to_update):
            if val:
                val[0].foo += data_to_update.foo
            else:
                # BAD THIS SHOULDVE TICKED WITH DATA_TO_SEND
                ...
        if val:
            csp.output(val[0])


class MyGetModule(GatewayModule):
    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        csp.add_graph_output("my_channel", channels.get_channel(MyGatewayChannels.my_channel))
        csp.add_graph_output("my_list_channel", channels.get_channel(MyGatewayChannels.my_list_channel))
        csp.add_graph_output("my_array_channel", channels.get_channel(MyGatewayChannels.my_array_channel))
        # Get by basket
        for k, v in channels.get_channel(MyGatewayChannels.my_enum_basket).items():
            csp.add_graph_output(f"my_enum_basket[{k}]", v)
        for k, v in channels.get_channel(MyGatewayChannels.my_str_basket).items():
            csp.add_graph_output(f"my_str_basket[{k}]", v)
        for k, v in channels.get_channel(MyGatewayChannels.my_enum_basket_list).items():
            csp.add_graph_output(f"my_enum_basket_list[{k}]", v)

        # Get by indexer
        csp.add_graph_output(
            "my_enum_basket_ONE",
            channels.get_channel(MyGatewayChannels.my_enum_basket, MyEnum.ONE),
        )
        csp.add_graph_output(
            "my_enum_basket_TWO",
            channels.get_channel(MyGatewayChannels.my_enum_basket, MyEnum.TWO),
        )
        csp.add_graph_output(
            "my_str_basket_my_key",
            channels.get_channel(MyGatewayChannels.my_str_basket, "my_key"),
        )
        csp.add_graph_output(
            "my_str_basket_my_key2",
            channels.get_channel(MyGatewayChannels.my_str_basket, "my_key2"),
        )


class MyGetModuleDynamicKeys(MyGetModule):
    requires: Optional[ChannelSelection] = []
    my_keys: List[str] = []

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: self.my_keys}


class MyFastShutdownModule(GatewayModule):
    shutdown_called: bool = False

    def connect(self, channels: MyGatewayChannels):
        return

    def shutdown(self):
        self.shutdown_called = True


class MySlowShutdownModule(GatewayModule):
    shutdown_called: bool = False

    def connect(self, channels: MyGatewayChannels):
        return

    def shutdown(self):
        time.sleep(10)


class MyInfiniteShutdownModule(GatewayModule):
    def connect(self, channels):
        return

    def shutdown(self):
        time.sleep(100000)


class MyGetModuleListRequires(MyGetModule):
    requires: Optional[ChannelSelection] = []


class MyGetModuleListRequiresAnnotated(MyGetModule):
    requires: Optional[ChannelSelection] = []


class MyBuildFailureModule(GatewayModule):
    def connect(self, channels: MyGatewayChannels) -> None:
        raise ValueError("Cannot build graph")


class MyAssertStartDetectorModule(GatewayModule):
    gateway: Dict[str, Gateway] = Field(default_factory=dict)

    @csp.node
    def _assert_started(self, timer: ts[bool]):
        with csp.state():
            # NOTE: The last csp.start() that runs will be the one that sets `running`
            assert self.gateway["value"].running is False
        if csp.ticked(timer):
            assert self.gateway["value"].running

    def connect(self, channels: MyGatewayChannels) -> None:
        self._assert_started(csp.const(True))


def test_run_no_modules():
    gateway = MyGateway(modules=[])
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


def test_user_modules():
    gateway = MyGateway(user_modules=[])
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


@pytest.mark.parametrize("by_key", [True, False])
def test_set_get(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[setter, getter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


def test_run_user_modules():
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    getter = MyGetModule()
    gateway = MyGateway(user_modules=[setter, getter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1


# TODO: We can't pass in a subclass to a channel that takes a specific class
def test_set_get_fails_with_subclass():
    setter = MySetModule(
        my_data=csp.const(MyChildStruct(foo=1.0, foo2=9)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[setter, getter], channels=MyGatewayChannels())
    with pytest.raises(TypeError):
        csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


@pytest.mark.parametrize("by_key", [True, False])
def test_get_set(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


@pytest.mark.parametrize("by_key", [False, True])
def test_get_set_block_set_past(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(
        modules=[getter, setter],
        channels=MyGatewayChannels(),
        block_set_channels_until=datetime(1969, 1, 2),
    )
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


@pytest.mark.parametrize("by_key", [False, True])
@pytest.mark.parametrize("override", [datetime(2020, 1, 1), None])
def test_get_set_block_set(by_key, override):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule(block_set_channels_until=override)
    gateway = MyGateway(
        modules=[getter, setter],
        channels=MyGatewayChannels(),
        block_set_channels_until=datetime(2020, 1, 2),
    )
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    for _, output in out.items():
        assert len(output) == 0


@pytest.mark.parametrize("by_key", [False, True])
def test_get_set_block_set_override(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
        block_set_channels_until=datetime(2020, 1, 1),
    )
    getter = MyGetModule(block_set_channels_until=datetime(2020, 1, 1))
    gateway = MyGateway(
        modules=[getter, setter],
        channels=MyGatewayChannels(),
        block_set_channels_until=datetime(2020, 1, 2),
    )
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


def test_get_module_dynamic_keys():
    """Set dynamic keys at the module level, and make sure building the channels works"""
    getter = MyGetModuleDynamicKeys(my_keys=["my_key", "my_key2"])
    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    for v in out.values():
        assert len(v) == 0


def test_get_channels_dynamic_keys():
    """Set dynamic keys at the channel level, and make sure building the channels works"""
    getter = MyGetModuleDynamicKeys()
    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels(my_keys=["my_key", "my_key2"]))
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    for v in out.values():
        assert len(v) == 0


def test_requires_default_conversion():
    getter = MyGetModuleDynamicKeys()
    assert isinstance(getter.requires, ChannelSelection)


def test_requires_fails():
    """Require a channel not set"""
    with pytest.raises(ValidationError):
        MyGetModuleDynamicKeys(requires=99)

    getter = MyGetModuleDynamicKeys(requires=dict(exclude=[MyGatewayChannels.my_array_channel]))
    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels(my_keys=["my_key", "my_key2"]))
    with pytest.raises(NoProviderException):
        csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


def test_requires_backwards_compatible():
    for GetterClass in [MyGetModuleListRequires, MyGetModuleListRequiresAnnotated]:
        getter = GetterClass()
        gateway = MyGateway(modules=[getter], channels=MyGatewayChannels(my_keys=["my_key", "my_key2"]))
        csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))

        getter = GetterClass(requires=[MyGatewayChannels.my_array_channel])
        gateway = MyGateway(modules=[getter], channels=MyGatewayChannels(my_keys=["my_key", "my_key2"]))
        with pytest.raises(NoProviderException):
            csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


@pytest.mark.parametrize("by_key", [True, False])
def test_get_set_set(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter, setter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 2
    assert len(out["my_list_channel"]) == 2
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_list_channel"][1][1]) == 2
    assert len(out["my_array_channel"]) == 2
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 2
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 2
    assert len(out["my_str_basket[my_key]"]) == 2
    assert len(out["my_str_basket[my_key2]"]) == 2
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 2
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 2
    assert len(out["my_enum_basket_ONE"]) == 2
    assert len(out["my_enum_basket_TWO"]) == 2
    assert len(out["my_str_basket_my_key"]) == 2
    assert len(out["my_str_basket_my_key2"]) == 2


@pytest.mark.parametrize("by_key", [True, False])
def test_get_set_set_block_set_on_one(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    setter2 = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
        block_set_channels_until=datetime(2020, 2, 1),
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter, setter2], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert len(out["my_array_channel"]) == 1
    assert len(out["my_array_channel"][0][1]) == 2
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket_ONE"]) == 1
    assert len(out["my_enum_basket_TWO"]) == 1
    assert len(out["my_str_basket_my_key"]) == 1
    assert len(out["my_str_basket_my_key2"]) == 1


def test_start_stop():
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter], channels=MyGatewayChannels())
    gateway.start(block=False, realtime=True, rest=False)
    gateway.stop()


@pytest.mark.parametrize("by_key", [True, False])
def test_last(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter], channels=MyGatewayChannels())
    gateway.start(block=False, realtime=True, rest=False)
    try:
        time.sleep(0.5)
        assert isinstance(gateway.channels.last("my_channel"), MyStruct)
        assert isinstance(gateway.channels.last("my_list_channel"), list)
        assert isinstance(gateway.channels.last("my_array_channel"), np.ndarray)

        output = gateway.channels.last("my_enum_basket")
        assert isinstance(output, dict)
        assert len(output) == 2
        output = gateway.channels.last("my_enum_basket", MyEnum.ONE)
        assert isinstance(output, MyStruct)

        output = gateway.channels.last("my_str_basket")
        assert isinstance(output, dict)
        assert len(output) == 2

        output = gateway.channels.last("my_str_basket", "my_key")
        assert isinstance(output, MyStruct)

        output = gateway.channels.last("my_enum_basket_list")
        assert isinstance(output, dict)
        assert len(output) == 2
    finally:
        gateway.stop()


@pytest.mark.parametrize("by_key", [True, False])
def test_state(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[getter, setter], channels=MyGatewayChannels())
    gateway.start(block=False, realtime=True, rest=False)
    try:
        time.sleep(0.5)
        state = gateway.channels.state("my_channel")
        assert isinstance(state, State)
        assert len(state.query()) == 1
        state = gateway.channels.state("my_list_channel")
        assert isinstance(state, State)
        assert len(state.query()) == 2
    finally:
        gateway.stop()


def test_build_failure():
    failure = MyBuildFailureModule()
    gateway = MyGateway(modules=[failure], channels=MyGatewayChannels())
    start = time.time()
    try:
        gateway.start(rest=True, _in_test=True, build_timeout=timedelta(seconds=10))
    except (SystemExit, RuntimeError):
        pass
    elapsed = time.time() - start
    assert elapsed < 5


def test_set_get_disable_set():
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    getter = MyGetModule(disable=True)
    gateway = MyGateway(modules=[setter, getter, setter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out is None


def test_set_get_set_disable_all():
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        disable=True,
    )
    getter = MyGetModule(disable=True)
    gateway = MyGateway(modules=[setter, getter, setter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out is None


def test_disable_get_after_module_initialization():
    getter = MyGetModule()
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    getter.disable = True
    assert getter.disable is True
    gateway = MyGateway(modules=[setter, getter, setter], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out is None


@pytest.mark.parametrize("by_key", [True, False])
def test_disable_fails_after_gateway_object_initialization(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    getter = MyGetModule()
    gateway = MyGateway(modules=[setter, getter, setter], channels=MyGatewayChannels())
    getter.disable = True
    assert getter.disable is True

    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
    assert out is None


def test_start_detector_order():
    """Make sure start detector is started before all other nodes."""
    m = MyAssertStartDetectorModule()
    gateway = MyGateway(modules=[m], channels=MyGatewayChannels())
    m.gateway["value"] = gateway
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))


def test_shutdown_fast():
    """Test whether the shutdown logic is invoked"""
    getter = MyFastShutdownModule()
    assert getter.shutdown_called is False
    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels())

    gateway.start(block=False, _in_test=True, realtime=True, rest=False)
    gateway.stop()

    assert getter.shutdown_called is True


def test_shutdown_slow():
    """Test whether the shutdown logic logs error"""
    root_logger = logging.getLogger()
    log_stream = StringIO()
    stream_handler = logging.StreamHandler(log_stream)
    root_logger.addHandler(stream_handler)
    getter = MySlowShutdownModule()
    assert getter.shutdown_called is False

    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels())
    gateway.start(block=False, _in_test=True, realtime=True, rest=False, module_shutdown_timeout=5)
    gateway.stop()
    res = log_stream.getvalue()
    assert "Shutting down modules took more than " in res
    assert "forcefully shutting down..." in res


def run_gateway():
    getter = MyInfiniteShutdownModule()
    gateway = MyGateway(modules=[getter], channels=MyGatewayChannels())
    gateway.start(block=False, _in_test=True, realtime=True, rest=False, module_shutdown_timeout=5)
    gateway.stop()


def test_shutdown_infinite():
    """Test whether gateway shutdown leads to infinitely waiting ThreadPool shutting down"""
    p = multiprocessing.Process(target=run_gateway)
    p.start()
    i = 0
    while p.is_alive() and i < 30:
        time.sleep(1)
        i += 1
    assert not p.is_alive()


class MySetModuleDynamicChannels(GatewayModule):
    scalar_channel_name: str
    list_channel_name: str
    connect_channels_assertion: Optional[Callable[[MyGatewayChannels], None]] = None

    def dynamic_channels(self) -> Optional[Dict[str, Union[Type[GatewayStruct], Type[List[GatewayStruct]]]]]:
        return {
            self.list_channel_name: List[MyStruct],
            self.scalar_channel_name: MyStruct,
        }

    def dynamic_state_channels(self) -> Optional[Set[str]]:
        return {self.scalar_channel_name, self.list_channel_name}

    def connect(self, channels: MyGatewayChannels) -> None:
        channels.set_channel(
            self.scalar_channel_name,
            csp.const(MyStruct(foo=1.0)),
        )
        channels.set_channel(
            self.list_channel_name,
            csp.const([MyStruct(foo=2.0), MyStruct(foo=3.0)]),
        )
        channels.set_state(self.scalar_channel_name, keyby="id")
        channels.set_state(self.list_channel_name, keyby="id")
        if self.connect_channels_assertion:
            self.connect_channels_assertion(channels)


class MyGetModuleDynamicChannels(GatewayModule):
    scalar_channel_name: str
    list_channel_name: str

    def connect(self, channels: MyGatewayChannels) -> None:
        csp.add_graph_output(
            self.scalar_channel_name,
            channels.get_channel(self.scalar_channel_name),
        )
        csp.add_graph_output(
            self.list_channel_name,
            channels.get_channel(self.list_channel_name),
        )
        csp.add_graph_output(
            f"s_{self.scalar_channel_name}",
            channels.get_channel(f"s_{self.scalar_channel_name}"),
        )
        csp.add_graph_output(
            f"s_{self.list_channel_name}",
            channels.get_channel(f"s_{self.list_channel_name}"),
        )


@pytest.mark.parametrize("gateway_from_config", (False, True))
@pytest.mark.parametrize("explicit_channels", (True, False))
def test_dynamic_channels(explicit_channels, gateway_from_config):
    if gateway_from_config:
        config = {
            "gateway": {
                "_target_": "csp_gateway.Gateway",
                "modules": ["/modules/setter", "/modules/getter"],
            },
            "modules": {
                "setter": {
                    "_target_": "csp_gateway.tests.server.gateway.test_gateway.MySetModuleDynamicChannels",
                    "scalar_channel_name": "my_dynamic_scalar_channel",
                    "list_channel_name": "my_dynamic_list_channel",
                },
                "getter": {
                    "_target_": "csp_gateway.tests.server.gateway.test_gateway.MyGetModuleDynamicChannels",
                    "scalar_channel_name": "my_dynamic_scalar_channel",
                    "list_channel_name": "my_dynamic_list_channel",
                },
            },
        }
        if explicit_channels:
            config["gateway"]["channels"] = {
                "_target_": "csp_gateway.tests.server.gateway.test_gateway.MyGatewayChannels",
            }

        config = OmegaConf.create(config)
        registry = ModelRegistry.root()
        registry.load_config(cfg=config, overwrite=True)
        gateway = registry["gateway"]

    else:
        setter = MySetModuleDynamicChannels(
            scalar_channel_name="my_dynamic_scalar_channel",
            list_channel_name="my_dynamic_list_channel",
        )
        getter = MyGetModuleDynamicChannels(
            scalar_channel_name="my_dynamic_scalar_channel",
            list_channel_name="my_dynamic_list_channel",
        )
        channels = MyGatewayChannels() if explicit_channels else None
        gateway = MyGateway(modules=[setter, getter], channels=channels)

    for _ in range(2):
        out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))
        assert len(out["my_dynamic_scalar_channel"]) == 1
        assert out["my_dynamic_scalar_channel"][0][1].foo == 1.0
        assert len(out["my_dynamic_list_channel"]) == 1
        assert out["my_dynamic_list_channel"][0][1][0].foo == 2.0
        assert out["my_dynamic_list_channel"][0][1][1].foo == 3.0
        assert len(out["s_my_dynamic_scalar_channel"]) == 1
        assert len(out["s_my_dynamic_list_channel"]) == 2
        assert out["my_dynamic_list_channel"][0][1][0].foo == 2.0
        assert out["my_dynamic_list_channel"][0][1][1].foo == 3.0


def test_conflicting_dynamic_channels():
    setter_1 = MySetModuleDynamicChannels(
        scalar_channel_name="channel_1",
        list_channel_name="channel_2",
    )
    setter_2 = MySetModuleDynamicChannels(
        scalar_channel_name="channel_2",
        list_channel_name="channel_1",
    )
    with pytest.raises(ValueError, match="Conflicting types for"):
        csp.build_graph(MyGateway(modules=[setter_1, setter_2], channels=MyGatewayChannels()).graph)


def test_dynamic_channels_same_static_properties():
    my_object = {}
    channel = MyGatewayChannels(my_static_dict_of_objects={"a": my_object})

    def assert_func(c: MyGatewayChannels):
        assert c.my_static_dict_of_objects["a"] is my_object

    setter = MySetModuleDynamicChannels(scalar_channel_name="channel_1", list_channel_name="channel_2", connect_channels_assertion=assert_func)
    csp.build_graph(MyGateway(modules=[setter], channels=channel).graph)


@pytest.mark.parametrize("harness_first", [False, True])
def test_harness_ordering(harness_first):
    set_get = MyStatefulGetSetModule(requires=[])
    CHANNELS = [
        MyGatewayChannels.my_channel_final,
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_channel_dup,
    ]
    h = GatewayTestHarness(test_channels=CHANNELS)
    h.send(
        MyGatewayChannels.my_channel,
        MyStruct(foo=1.0),
    )
    h.send(MyGatewayChannels.my_channel_dup, MyStruct(foo=1.0))
    """
    OK, so the expected behavior, is that since we dont have any timers or alarms
    And there is no unrolling, everything should tick the same engine cycle.

    If everything ticks the same engine cycle, as expected, then foo should be 2.0
    Previously, depending on the ordering of modules, sometimes "my_channel_final" had "foo" with value of 1.0
    """
    h.assert_ticked(MyGatewayChannels.my_channel_final, 1)
    h.assert_attr_equal(MyGatewayChannels.my_channel_final, "foo", 2.0)
    if harness_first:
        gateway = MyGateway(
            modules=[h, set_get],
            channels=MyGatewayChannels(),
        )
    else:
        gateway = MyGateway(
            modules=[set_get, h],
            channels=MyGatewayChannels(),
        )
    csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(1))



from csp_gateway import GatewayModule, get_counter


class MyGatewayModule(GatewayModule):
    def connect(self, channels):
        pass

    def shutdown(self):
        pass


class IDGenerator:
    def time_get_counter(self):
        _ = get_counter(MyGatewayModule())


class Counter:
    def setup(self):
        self.counter = get_counter(MyGatewayModule())

    def time_counter_next(self):
        self.counter.next()



import logging
import time
from datetime import timedelta
from typing import Any, Callable, Optional

import pandas as pd
import pyarrow as pa
from csp import ts
from csp.impl.pushadapter import PushInputAdapter
from csp.impl.wiring import py_push_adapter_def

from csp_gateway.utils import get_thread

__all__ = (
    "sql_polling_adapter_def",
    "poll_sql_for_arrow_tbl",
    "poll_sql_for_pandas_df",
)


def poll_sql_for_arrow_tbl(connection: str, query: str, logger_name: str = __name__) -> pa.Table:
    from arrow_odbc import read_arrow_batches_from_odbc

    reader = read_arrow_batches_from_odbc(
        query=query,
        connection_string=connection,
        batch_size=10_000,
    )
    return pa.Table.from_batches(batches=reader, schema=reader.schema)


def poll_sql_for_pandas_df(connection: str, query: str, logger_name: str = __name__) -> pd.DataFrame:
    return poll_sql_for_arrow_tbl(connection, query, logger_name).to_pandas()


class PollingSQLAdapterImpl(PushInputAdapter):
    def __init__(
        self,
        interval: timedelta,
        connection: str,
        query: str,
        poll: Callable[[str, str, logging.Logger], Any],
        callback: Optional[Callable[[Any], Any]] = None,
        logger_name: str = __name__,
        failed_poll_msg: str = "Failed to poll sql database",
        connection_timeout_seconds: int = 0,
        query_timeout_seconds: int = 0,
    ):
        self._interval = interval
        self._connection = connection
        if connection_timeout_seconds:
            self._connection += f";timeout={connection_timeout_seconds}"

        if query_timeout_seconds:
            self._connection += f";commandTimeout={query_timeout_seconds}"

        self._query = query
        self._callback = callback
        self._last_update = 0
        self._poll = poll
        self._logger_name = logger_name
        self._failed_poll_msg = failed_poll_msg

        self._thread = None
        self._running = False
        self._paused = False

    def start(self, starttime, endtime):
        """start will get called at the start of the engine, at which point the push
        input adapter should start its thread that will push the data onto the adapter.  Note
        that push adapters will ALWAYS have a separate thread driving ticks into the csp engine thread
        """
        self._running = True
        self._thread = get_thread(target=self._run)
        self._thread.start()

    def stop(self):
        """stop will get called at the end of the run, at which point resources should
        be cleaned up
        """
        if self._running:
            self._running = False
            # if it is paused, we just kill it
            if self._paused:
                self._thread.join(0.01)
            else:
                # we give it the interval time to finish
                self._thread.join(self._interval.total_seconds())

    def _run(self):
        log = logging.getLogger(self._logger_name)
        while self._running:
            self._paused = False
            try:
                now = time.time()
                res = self._poll(self._connection, self._query, self._logger_name)
                if self._callback is not None:
                    res = self._callback(res)
                self._last_update = now
                self.push_tick(res)

            except Exception:
                # TODO: What do we want to do here?
                log.warning(self._failed_poll_msg, exc_info=True)
                self.push_tick(None)  # None is a bad value

            # TODO: Handle what happens if it hangs, by having a separate thread listen for data in a queue
            #  and publish pd.DataFrame if it misses a heartbeat
            #  sleep interval
            self._paused = True
            time.sleep(self._interval.total_seconds())


sql_polling_adapter_def = py_push_adapter_def(
    "sql_polling_adapter_def",
    PollingSQLAdapterImpl,
    ts[object],
    interval=timedelta,
    connection=str,
    query=str,
    poll=object,
    callback=object,
    logger_name=str,
    failed_poll_msg=str,
    connection_timeout_seconds=int,
    query_timeout_seconds=int,
)



from typing import Optional

import csp
from pydantic import Field

from csp_gateway.server import ChannelSelection, ChannelsType, GatewayModule


class PrintChannels(GatewayModule):
    """
    Gateway Module for printing channels, which could be useful for debugging.
    There exists a designated logging node class `LogChannels` which is preferred
    and can specify a logger to use by name.
    """

    selection: ChannelSelection = Field(default_factory=ChannelSelection)
    requires: Optional[ChannelSelection] = []

    def connect(self, channels: ChannelsType):
        for field in self.selection.select_from(channels):
            csp.print(f"{field}", channels.get_channel(field))



from datetime import datetime, timedelta, timezone
from typing import Annotated, Any, Dict, List, Optional, Tuple, Type

import csp
import orjson
import pytest
from csp import ts
from pydantic import GetPydanticSchema, PrivateAttr
from pydantic_core import core_schema

from csp_gateway import (
    AddChannelsToGraphOutput,
    Channels,
    ChannelSelection,
    Gateway,
    GatewayChannels,
    GatewayModule,
    GatewayStruct,
    JSONConverter,
)
from csp_gateway.server.gateway.csp.channels import _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD
from csp_gateway.server.shared.json_converter import (
    ChannelValueModel as CVM,
    _convert_orjson_compatible,
    _create_snapshot_dict,
)
from csp_gateway.testing.shared_helpful_classes import (
    MyEnum,
    MyGateway,
    MyGatewayChannels,
    MySetModule,
    MyStruct,
)


class MysteryClass: ...


# This is an example on how to annotate a new, custom class, for validation and ser/der via pydantic
# GatewayStructs will pick this up and allow users to use their own defined serializtion and deserialization
ValidatedMysteryClass = Annotated[
    MysteryClass,  # our own special class Pydantic is not aware of
    GetPydanticSchema(
        lambda tp, handler: core_schema.no_info_before_validator_function(
            lambda x: MysteryClass() if x == "A" else "SHOULD_NEVER_HIT",
            schema=core_schema.any_schema(),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda x: "A", info_arg=False, return_schema=core_schema.str_schema(), when_used="json"
            ),  # this defines how to serialize the class. This will be used for serialization to str, such as for saving to json
        )  # outer level definition of schema, here, we say that "the schema can be anything, but before you apply the schema, run this validator function"
    ),
]


class MyPrivateAttrStruct(MyStruct):
    _private: str = "unset"


class MySelectiveSetModule(GatewayModule):
    requires: Optional[ChannelSelection] = []
    my_data: ts[MyStruct]
    my_data2: ts[MyStruct]
    my_list_data: ts[List[MyStruct]]
    by_key: bool = True

    def dynamic_keys(self):
        return {MyGatewayChannels.my_str_basket: ["my_key", "my_key2"]}

    def connect(self, channels: MyGatewayChannels) -> None:
        # Build a set of inputs based on the test cases
        channels.set_channel(MyGatewayChannels.my_channel, self.my_data)
        channels.set_channel(
            MyGatewayChannels.my_list_channel,
            self.my_list_data,
        )
        channels.add_send_channel(MyGatewayChannels.my_channel)
        channels.set_state(MyGatewayChannels.my_channel, "id")
        channels.add_send_channel(MyGatewayChannels.my_list_channel)
        channels.set_state(MyGatewayChannels.my_list_channel, "id")

        if self.by_key:
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data, MyEnum.ONE)
            channels.set_channel(MyGatewayChannels.my_enum_basket, self.my_data2, MyEnum.TWO)
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data, "my_key")
            channels.set_channel(MyGatewayChannels.my_str_basket, self.my_data2, "my_key2")
        else:
            channels.set_channel(
                MyGatewayChannels.my_enum_basket,
                {
                    MyEnum.ONE: self.my_data,
                    MyEnum.TWO: self.my_data2,
                },
            )
            channels.set_channel(
                MyGatewayChannels.my_str_basket,
                {
                    "my_key": self.my_data,
                    "my_key2": self.my_data2,
                },
            )


class MyCustomStruct(GatewayStruct):
    mystery_val: List[ValidatedMysteryClass]


class MyFatPipeChannels(MyGatewayChannels):
    fat_pipe: ts[str] = None
    custom_struct: ts[MyCustomStruct] = None


class SetMysteryClassModule(GatewayModule):
    custom_struct: ts[MyCustomStruct]

    def connect(self, channels):
        channels.set_channel(MyFatPipeChannels.custom_struct, self.custom_struct)


class MyFatPipeChannelsGateway(Gateway):
    channels_model: Type[Channels] = MyFatPipeChannels


class MyFatPipeSetter(GatewayModule):
    to_fat_pipe: ts[str]

    def connect(self, channels: MyFatPipeChannels):
        channels.set_channel(MyFatPipeChannels.fat_pipe, self.to_fat_pipe)


class MyJsonEncoder(GatewayModule):
    requires: Optional[ChannelSelection] = []
    channels_list: List[str]
    _dict_basket_keys: Dict[str, Any] = PrivateAttr(default=None)

    def dynamic_keys(self):
        self._dict_basket_keys = {
            MyGatewayChannels.my_str_basket: ["my_key", "my_key2"],
            MyGatewayChannels.my_enum_basket: list(MyEnum),
            MyGatewayChannels.my_enum_basket_list: list(MyEnum),
        }
        return self._dict_basket_keys

    def connect(self, channels: MyGatewayChannels):
        json_connector = JSONConverter(
            decode_channels=self.channels_list,
            encode_channels=self.channels_list,
            channels=channels,
        )

        csp.add_graph_output("encoded_snapshot_model", json_connector.encode())


class MyJsonDecoder(GatewayModule):
    requires: Optional[ChannelSelection] = []
    channels_list: List[str]
    _dict_basket_keys: Dict[str, Any] = PrivateAttr(default=None)
    flag_updates: Dict[str, List[Tuple[str, bool]]] = None

    def dynamic_keys(self):
        self._dict_basket_keys = {
            MyFatPipeChannels.my_str_basket: ["my_key", "my_key2"],
            MyFatPipeChannels.my_enum_basket: list(MyEnum),
            MyFatPipeChannels.my_enum_basket_list: list(MyEnum),
        }
        return self._dict_basket_keys

    def connect(self, channels: MyFatPipeChannels):
        json_connector = JSONConverter(
            decode_channels=self.channels_list,
            encode_channels=self.channels_list,
            channels=channels,
            flag_updates=self.flag_updates or {},
        )
        json_connector.decode(channels.get_channel(MyFatPipeChannels.fat_pipe))


class MyJsonEncoderDecoder(GatewayModule):
    requires: Optional[ChannelSelection] = []
    decode_channels: List[str]
    encode_channels: List[str]
    _dict_basket_keys: Dict[str, Any] = PrivateAttr(default=None)
    flag_updates: Dict[str, List[Tuple[str, bool]]] = None
    encode: bool

    def dynamic_keys(self):
        self._dict_basket_keys = {
            MyFatPipeChannels.my_str_basket: ["my_key", "my_key2"],
            MyFatPipeChannels.my_enum_basket: list(MyEnum),
            MyFatPipeChannels.my_enum_basket_list: list(MyEnum),
        }
        return self._dict_basket_keys

    def connect(self, channels: MyFatPipeChannels):
        json_connector = JSONConverter(
            decode_channels=self.decode_channels,
            encode_channels=self.encode_channels,
            channels=channels,
            flag_updates=self.flag_updates or {},
        )
        if self.encode:
            csp.add_graph_output("encoded_snapshot_model", json_connector.encode())
        else:
            json_connector.decode(channels.get_channel(MyFatPipeChannels.fat_pipe))


class MyJsonChecker(GatewayModule):
    requires: Optional[ChannelSelection] = []
    timestamp: datetime

    @csp.node
    def check_time(self, t: ts[datetime]) -> ts[bool]:
        if csp.ticked(t):
            return t == self.timestamp

    def connect(self, channels: MyFatPipeChannels):
        struct = channels.get_channel(MyFatPipeChannels.my_channel)
        csp.add_graph_output("engine_times_align", self.check_time(csp.times(struct)))


# This function is tested and used in other tests so it was factored out
def create_snapshot_dict(timestamp, dummy_id):
    channel_values_list = [
        CVM(
            channel=MyGatewayChannels.my_channel,
            value=MyStruct(foo=3.0, timestamp=timestamp, id=dummy_id),
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_list_channel,
            value=[
                MyStruct(foo=4.0, timestamp=timestamp, id=dummy_id),
                MyStruct(foo=5.0, timestamp=timestamp, id=dummy_id),
            ],
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_enum_basket,
            value=MyStruct(foo=6.0, timestamp=timestamp, id=dummy_id),
            dict_basket_key=MyEnum.ONE,
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_enum_basket,
            value=MyStruct(foo=7.0, timestamp=timestamp, id=dummy_id),
            dict_basket_key=MyEnum.TWO,
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_str_basket,
            value=MyStruct(foo=8.0, timestamp=timestamp, id=dummy_id),
            dict_basket_key="key1",
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_str_basket,
            value=MyStruct(foo=9.0, timestamp=timestamp, id=dummy_id),
            dict_basket_key="key2",
            timestamp=timestamp,
        ),
        CVM(
            channel=MyGatewayChannels.my_enum_basket_list,
            value=[
                MyStruct(foo=10.0, timestamp=timestamp, id=dummy_id),
                MyStruct(foo=11.0, timestamp=timestamp, id=dummy_id),
            ],
            dict_basket_key=MyEnum.ONE,
            timestamp=timestamp,
        ),
    ]
    snapshot_dict = _create_snapshot_dict(channel_values_list)
    return snapshot_dict


def test_convert_orjson_compatible():
    gateway_struct_list = [MyStruct(foo=3.0, time=timedelta(days=4, microseconds=11)), MyStruct(foo=2.9)]
    pydantic_list = _convert_orjson_compatible(gateway_struct_list)
    for pydantic_x, x in zip(pydantic_list, gateway_struct_list):
        target = x.type_adapter().dump_python(x, mode="json")
        target["timestamp"] = datetime.fromisoformat(target["timestamp"])
        assert pydantic_x == target
    # To avoid json encoding issues with these custom enum types
    # _convert_orjson_compatible converts them to their name for encoding
    enum = MyEnum.ONE
    assert _convert_orjson_compatible(enum) == enum.name

    my_str = "hai"
    assert _convert_orjson_compatible(my_str) == my_str


def test_parse_snapshot_dict():
    timestamp = datetime(2020, 1, 1)
    dummy_id = "9"
    snapshot_dict = create_snapshot_dict(timestamp, dummy_id)
    decoded_obj = MyGatewayChannels._snapshot_model.model_validate(snapshot_dict)

    assert decoded_obj.my_list_channel == [
        MyStruct(foo=4.0, timestamp=timestamp, id=dummy_id),
        MyStruct(foo=5.0, timestamp=timestamp, id=dummy_id),
    ]
    assert decoded_obj.my_channel == MyStruct(foo=3.0, timestamp=timestamp, id=dummy_id)
    assert decoded_obj.my_enum_basket[MyEnum.ONE] == MyStruct(foo=6.0, timestamp=timestamp, id=dummy_id)
    assert decoded_obj.my_enum_basket[MyEnum.TWO] == MyStruct(foo=7.0, timestamp=timestamp, id=dummy_id)
    assert decoded_obj.my_str_basket["key1"] == MyStruct(foo=8.0, timestamp=timestamp, id=dummy_id)
    assert decoded_obj.my_str_basket["key2"] == MyStruct(foo=9.0, timestamp=timestamp, id=dummy_id)
    assert decoded_obj.my_enum_basket_list[MyEnum.ONE] == [
        MyStruct(foo=10.0, timestamp=timestamp, id=dummy_id),
        MyStruct(foo=11.0, timestamp=timestamp, id=dummy_id),
    ]


def test_parse_snapshot_dict_with_private_fields():
    class MyPrivateAttrGatewayChannels(GatewayChannels):
        my_channel: ts[MyPrivateAttrStruct] = None

    channel_values_list = [
        CVM(
            channel=MyPrivateAttrGatewayChannels.my_channel,
            value=MyPrivateAttrStruct(foo=3.0, _private="howdy"),
            timestamp=datetime(2020, 1, 1),
        )
    ]
    snapshot_dict = _create_snapshot_dict(channel_values_list)
    assert "_private" not in snapshot_dict[MyPrivateAttrGatewayChannels.my_channel]

    # When we decode to csp, the underscore field comes back and is set to its default value
    snapshot_model = MyPrivateAttrGatewayChannels._snapshot_model.model_validate(snapshot_dict)
    assert snapshot_model.my_channel._private == "unset"


# NOTE: This test relies on the format of the json encoding to not change
def test_json_snapshot():
    timestamp = datetime(2020, 1, 1)
    dummy_id = "9"
    snapshot_dict = create_snapshot_dict(timestamp, dummy_id)
    model = MyGatewayChannels._snapshot_model.model_validate(snapshot_dict)
    target = '{"my_channel":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":3.0,"my_flag":true},"my_list_channel":[{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":4.0,"my_flag":true},{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":5.0,"my_flag":true}],"my_enum_basket":{"ONE":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":6.0,"my_flag":true},"TWO":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":7.0,"my_flag":true}},"my_str_basket":{"key1":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":8.0,"my_flag":true},"key2":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":9.0,"my_flag":true}},"my_enum_basket_list":{"ONE":[{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":10.0,"my_flag":true},{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":11.0,"my_flag":true}]},"csp_engine_timestamp":"2020-01-01T00:00:00+00:00"}'  # noqa
    target = target.replace("+00:00", "")
    print(model.model_dump_json())
    assert orjson.loads(model.model_dump_json()) == orjson.loads(target)


# NOTE: This test relies on the format of the json encoding to not change
def test_parse_snapshot_json():
    timestamp = datetime(2020, 1, 1, 5, tzinfo=timezone(timedelta(hours=5)))
    dummy_id = "9"
    snapshot_str = '{"my_channel":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":3.0},"my_list_channel":[{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":4.0},{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":5.0}],"my_enum_basket":{"ONE":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":6.0},"TWO":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":7.0}},"my_str_basket":{"key1":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":8.0},"key2":{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":9.0}},"my_enum_basket_list":{"ONE":[{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":10.0},{"id":"9","timestamp":"2020-01-01T00:00:00+00:00","foo":11.0}]}}'  # noqa: E501
    model = MyGatewayChannels._snapshot_model.model_validate_json(snapshot_str)
    exp_model = MyGatewayChannels._snapshot_model(
        my_channel=MyStruct(foo=3.0, timestamp=timestamp, id=dummy_id),
        my_list_channel=[
            MyStruct(foo=4.0, timestamp=timestamp, id=dummy_id),
            MyStruct(foo=5.0, timestamp=timestamp, id=dummy_id),
        ],
        my_enum_basket={
            MyEnum.ONE: MyStruct(foo=6.0, timestamp=timestamp, id=dummy_id),
            MyEnum.TWO: MyStruct(foo=7.0, timestamp=timestamp, id=dummy_id),
        },
        my_str_basket={
            "key1": MyStruct(foo=8.0, timestamp=timestamp, id=dummy_id),
            "key2": MyStruct(foo=9.0, timestamp=timestamp, id=dummy_id),
        },
        my_enum_basket_list={
            MyEnum.ONE: [
                MyStruct(foo=10.0, timestamp=timestamp, id=dummy_id),
                MyStruct(foo=11.0, timestamp=timestamp, id=dummy_id),
            ]
        },
    )
    # The models align once we get back to the csp version of the structs
    assert model.my_channel == exp_model.my_channel
    assert model.my_list_channel == exp_model.my_list_channel
    assert model.my_str_basket["key1"] == exp_model.my_str_basket["key1"]
    assert model.my_str_basket["key2"] == exp_model.my_str_basket["key2"]
    assert model.my_enum_basket[MyEnum.ONE] == exp_model.my_enum_basket[MyEnum.ONE]
    assert model.my_enum_basket[MyEnum.TWO] == exp_model.my_enum_basket[MyEnum.TWO]
    assert model.my_enum_basket_list[MyEnum.ONE] == exp_model.my_enum_basket_list[MyEnum.ONE]


@pytest.mark.parametrize("by_key", [True, False])
def test_encode(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0, time=timedelta(days=4, microseconds=11))),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]
    snapshot_model = MyGatewayChannels._snapshot_model.model_validate_json(encoded_snapshot_model.encoding)
    assert snapshot_model.my_channel.foo == 1.0
    assert snapshot_model.my_channel.time == timedelta(days=4, microseconds=11)
    assert snapshot_model.my_list_channel[1].foo == 2.0
    assert snapshot_model.my_str_basket["my_key"].foo == 1.0
    assert snapshot_model.my_str_basket["my_key2"].foo == 2.0
    assert snapshot_model.my_enum_basket[MyEnum.ONE].foo == 1.0
    assert snapshot_model.my_enum_basket[MyEnum.TWO].foo == 2.0
    assert snapshot_model.my_enum_basket_list[MyEnum.ONE] == snapshot_model.my_enum_basket_list[MyEnum.TWO]


@pytest.mark.parametrize("by_key", [True, False])
def test_encode_filter_channels(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    channels = list(MyGatewayChannels.model_fields.keys())
    assert MyGatewayChannels.s_my_channel in channels
    assert MyGatewayChannels.my_array_channel in channels
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]
    snapshot_model = MyGatewayChannels._snapshot_model.model_validate_json(encoded_snapshot_model.encoding)
    assert snapshot_model.my_channel.foo == 1.0
    assert snapshot_model.my_list_channel[1].foo == 2.0
    assert snapshot_model.my_str_basket["my_key"].foo == 1.0
    assert snapshot_model.my_str_basket["my_key2"].foo == 2.0
    assert snapshot_model.my_enum_basket[MyEnum.ONE].foo == 1.0
    assert snapshot_model.my_enum_basket[MyEnum.TWO].foo == 2.0
    assert snapshot_model.my_enum_basket_list[MyEnum.ONE] == snapshot_model.my_enum_basket_list[MyEnum.TWO]
    assert MyGatewayChannels.s_my_channel not in snapshot_model.model_fields
    assert getattr(snapshot_model, _CSP_ENGINE_CYCLE_TIMESTAMP_FIELD) == datetime(2020, 1, 1)


@pytest.mark.parametrize("by_key", [True, False])
def test_encode_twice_same_timestamp(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=3.0), MyStruct(foo=9.0)]),
        by_key=by_key,
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    assert len(out) == 1
    assert len(out["encoded_snapshot_model"]) == 2
    assert out["encoded_snapshot_model"][0][0] == out["encoded_snapshot_model"][1][0]

    for _, encoding in out["encoded_snapshot_model"]:
        snapshot_model = MyGatewayChannels._snapshot_model.model_validate_json(encoding.encoding)
        assert snapshot_model.my_channel.foo == 1.0
        assert snapshot_model.my_list_channel[1].foo == 9.0
        assert snapshot_model.my_str_basket["my_key"].foo == 1.0
        assert snapshot_model.my_str_basket["my_key2"].foo == 2.0
        assert snapshot_model.my_enum_basket[MyEnum.ONE].foo == 1.0
        assert snapshot_model.my_enum_basket[MyEnum.TWO].foo == 2.0
        assert snapshot_model.my_enum_basket_list[MyEnum.ONE] == snapshot_model.my_enum_basket_list[MyEnum.TWO]


@pytest.mark.parametrize("by_key", [True, False])
def test_encode_decode(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    raw_custom_struct = MyCustomStruct.type_adapter().validate_python(dict(mystery_val=["A", "A"]))  # must be A for serialization
    setter_mystery = SetMysteryClassModule(custom_struct=csp.const(raw_custom_struct))
    channels = [
        MyFatPipeChannels.my_channel,
        MyFatPipeChannels.my_list_channel,
        MyFatPipeChannels.my_str_basket,
        MyFatPipeChannels.my_enum_basket,
        MyFatPipeChannels.my_enum_basket_list,
        MyFatPipeChannels.custom_struct,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder, setter_mystery], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]

    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(encoded_snapshot_model.encoding))
    getter = AddChannelsToGraphOutput(selection=channels)
    json_decoder = MyJsonDecoder(channels_list=channels)
    gateway = MyFatPipeChannelsGateway(modules=[fat_pipe_module, json_decoder, getter], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=5))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert out["my_list_channel"][0][1][1].foo == 2.0
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert isinstance(out["custom_struct"][0][1].mystery_val[0], MysteryClass)  # custom serialization worked!
    assert len(out["custom_struct"][0][1].mystery_val) == 2


@pytest.mark.parametrize("delay", [timedelta(seconds=1), timedelta(seconds=5)])
def test_encode_decode_timestamp(delay):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0), delay=delay),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][1][1]

    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(encoded_snapshot_model.encoding))
    getter = AddChannelsToGraphOutput(selection=channels)
    json_decoder = MyJsonDecoder(channels_list=channels)
    json_checker = MyJsonChecker(timestamp=datetime(2020, 1, 1) + delay)
    gateway = MyFatPipeChannelsGateway(
        modules=[fat_pipe_module, json_decoder, getter, json_checker],
        channels=MyFatPipeChannels(),
    )
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=6))
    assert out["engine_times_align"] == [(datetime(2020, 1, 1) + delay, True)]


@pytest.mark.parametrize("by_key", [True, False])
def test_encode_decode_missing_channels(by_key):
    setter = MySelectiveSetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]

    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(encoded_snapshot_model.encoding))
    getter = AddChannelsToGraphOutput(selection=channels)
    json_decoder = MyJsonDecoder(channels_list=channels)
    gateway = MyFatPipeChannelsGateway(modules=[fat_pipe_module, json_decoder, getter], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=5))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert out["my_list_channel"][0][1][1].foo == 2.0
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 0
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 0
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1


@pytest.mark.parametrize("by_key", [True, False])
def test_encode_decode_some_channels(by_key):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
        by_key=by_key,
    )
    encode_channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    decode_channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
    ]
    json_encoder = MyJsonEncoderDecoder(encode_channels=encode_channels, decode_channels=decode_channels, encode=True)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]

    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(encoded_snapshot_model.encoding))
    getter = AddChannelsToGraphOutput(selection=encode_channels)
    json_decoder = MyJsonEncoderDecoder(encode_channels=encode_channels, decode_channels=decode_channels, encode=False)
    gateway = MyFatPipeChannelsGateway(modules=[fat_pipe_module, json_decoder, getter], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=5))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert out["my_list_channel"][0][1][1].foo == 2.0
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 0
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 0
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1


@pytest.mark.parametrize("include_list_channel", [True, False])
def test_encode_decode_flag_updates(include_list_channel):
    setter = MySetModule(
        my_data=csp.const(MyStruct(foo=1.0)),
        my_data2=csp.const(MyStruct(foo=2.0)),
        my_list_data=csp.const([MyStruct(foo=1.0), MyStruct(foo=2.0)]),
    )
    channels = [
        MyGatewayChannels.my_channel,
        MyGatewayChannels.my_list_channel,
        MyGatewayChannels.my_str_basket,
        MyGatewayChannels.my_enum_basket,
        MyGatewayChannels.my_enum_basket_list,
    ]
    json_encoder = MyJsonEncoder(channels_list=channels)
    gateway = MyGateway(modules=[setter, json_encoder], channels=MyGatewayChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1))
    encoded_snapshot_model = out["encoded_snapshot_model"][0][1]

    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(encoded_snapshot_model.encoding))
    getter = AddChannelsToGraphOutput(selection=channels)
    if include_list_channel:
        json_decoder = MyJsonDecoder(
            channels_list=channels,
            flag_updates={
                "my_channel": [("_new", False), ("my_flag", False)],
                "my_list_channel": [("_new", False)],
            },
        )
    else:
        json_decoder = MyJsonDecoder(
            channels_list=channels,
            flag_updates={"my_channel": [("_new", False), ("my_flag", False)]},
        )
    gateway = MyFatPipeChannelsGateway(modules=[fat_pipe_module, json_decoder, getter], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=5))
    assert len(out["my_channel"]) == 1
    assert len(out["my_list_channel"]) == 1
    assert len(out["my_list_channel"][0][1]) == 2
    assert out["my_list_channel"][0][1][1].foo == 2.0
    assert out["my_channel"][0][1]._new is False
    assert out["my_channel"][0][1].my_flag is False
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
    assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
    assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
    assert len(out["my_str_basket[my_key]"]) == 1
    assert len(out["my_str_basket[my_key2]"]) == 1


# NOTE: This test relies on the format of the json encoding to not change
@pytest.mark.parametrize("after_graph_ends", [True, False])
def test_decode(after_graph_ends):
    raw_encoding = """{"csp_engine_timestamp":"2020-01-01T00:00:00.000000+00:00","my_channel":{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},"my_list_channel":[{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},{"id":"2319507561549660169","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":2.0}],"my_enum_basket":{"ONE":{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},"TWO":{"id":"2319507561549660169","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":2.0}},"my_str_basket":{"my_key":{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},"my_key2":{"id":"2319507561549660169","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":2.0}},"my_enum_basket_list":{"TWO":[{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},{"id":"2319507561549660169","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":2.0}],"ONE":[{"id":"2319507561549660168","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":1.0},{"id":"2319507561549660169","timestamp":"2023-03-28T14:11:46.074000+00:00","foo":2.0}]}}"""  # noqa: E501

    if after_graph_ends:
        delay = timedelta(seconds=100)
    else:
        delay = timedelta()
    fat_pipe_module = MyFatPipeSetter(to_fat_pipe=csp.const(raw_encoding, delay=delay))
    channels = [
        MyFatPipeChannels.my_channel,
        MyFatPipeChannels.my_list_channel,
        MyFatPipeChannels.my_str_basket,
        MyFatPipeChannels.my_enum_basket,
        MyFatPipeChannels.my_enum_basket_list,
    ]
    getter = AddChannelsToGraphOutput(selection=channels)
    json_decoder = MyJsonDecoder(channels_list=channels)
    gateway = MyFatPipeChannelsGateway(modules=[fat_pipe_module, json_decoder, getter], channels=MyFatPipeChannels())
    out = csp.run(gateway.graph, starttime=datetime(2020, 1, 1), endtime=timedelta(seconds=5))

    if after_graph_ends:
        for v in out.values():
            assert not v

    else:
        assert len(out["my_channel"]) == 1
        assert len(out["my_list_channel"]) == 1
        assert len(out["my_list_channel"][0][1]) == 2
        assert out["my_list_channel"][0][1][1].foo == 2.0
        assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
        assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
        assert len(out["my_str_basket[my_key]"]) == 1
        assert len(out["my_str_basket[my_key2]"]) == 1
        assert len(out["my_enum_basket_list[MyEnum.ONE]"]) == 1
        assert len(out["my_enum_basket_list[MyEnum.TWO]"]) == 1
        assert len(out["my_enum_basket[MyEnum.ONE]"]) == 1
        assert len(out["my_enum_basket[MyEnum.TWO]"]) == 1
        assert len(out["my_str_basket[my_key]"]) == 1
        assert len(out["my_str_basket[my_key2]"]) == 1



from datetime import timedelta

from fastapi import APIRouter, Depends, HTTPException, Request, Security
from fastapi.responses import HTMLResponse, JSONResponse, RedirectResponse
from pydantic import Field, PrivateAttr
from starlette.status import HTTP_403_FORBIDDEN

from csp_gateway.server import GatewayChannels, GatewayModule

# separate to avoid circular
from ..web import GatewayWebApp
from .hacks.api_key_middleware_websocket_fix.api_key import (
    APIKeyCookie,
    APIKeyHeader,
    APIKeyQuery,
)


class MountAPIKeyMiddleware(GatewayModule):
    api_key_timeout: timedelta = Field(description="Cookie timeout for API Key authentication", default=timedelta(hours=12))

    # NOTE: don't make this publically configureable
    # as it is needed in gateway.py
    _api_key_name: str = PrivateAttr("token")
    _api_key_secret: str = PrivateAttr("")

    unauthorized_status_message: str = "unauthorized"

    def connect(self, channels: GatewayChannels) -> None:
        # NO-OP
        ...

    def rest(self, app: GatewayWebApp) -> None:
        if app.settings.AUTHENTICATE:
            # first, pull out the api key secret from the settings
            self._api_key_secret = app.settings.API_KEY

            # reinitialize header
            api_key_query = APIKeyQuery(name=self._api_key_name, auto_error=False)
            api_key_header = APIKeyHeader(name=self._api_key_name, auto_error=False)
            api_key_cookie = APIKeyCookie(name=self._api_key_name, auto_error=False)

            # routers
            auth_router: APIRouter = app.get_router("auth")
            public_router: APIRouter = app.get_router("public")

            # now mount middleware
            async def get_api_key(
                api_key_query: str = Security(api_key_query),
                api_key_header: str = Security(api_key_header),
                api_key_cookie: str = Security(api_key_cookie),
            ):
                if api_key_query == self._api_key_secret or api_key_header == self._api_key_secret or api_key_cookie == self._api_key_secret:
                    return self._api_key_secret
                else:
                    raise HTTPException(
                        status_code=HTTP_403_FORBIDDEN,
                        detail=self.unauthorized_status_message,
                    )

            @auth_router.get("/login")
            async def route_login_and_add_cookie(api_key: str = Depends(get_api_key)):
                response = RedirectResponse(url="/")
                response.set_cookie(
                    self._api_key_name,
                    value=api_key,
                    domain=app.settings.AUTHENTICATION_DOMAIN,
                    httponly=True,
                    max_age=self.api_key_timeout.total_seconds(),
                    expires=self.api_key_timeout.total_seconds(),
                )
                return response

            @auth_router.get("/logout")
            async def route_logout_and_remove_cookie():
                response = RedirectResponse(url="/login")
                response.delete_cookie(self._api_key_name, domain=app.settings.AUTHENTICATION_DOMAIN)
                return response

            # I'm hand rolling these for now...
            @public_router.get("/login", response_class=HTMLResponse, include_in_schema=False)
            async def get_login_page(token: str = "", request: Request = None):
                if token:
                    if token != "":
                        return RedirectResponse(url=f"{app.settings.API_STR}/auth/login?token={token}")
                return app.templates.TemplateResponse(
                    "login.html.j2",
                    {"request": request, "api_key_name": self._api_key_name},
                )

            @public_router.get("/logout", response_class=HTMLResponse, include_in_schema=False)
            async def get_logout_page(request: Request = None):
                return app.templates.TemplateResponse("logout.html.j2", {"request": request})

            # add auth to all other routes
            app.add_middleware(Depends(get_api_key))

            @app.app.exception_handler(403)
            async def custom_403_handler(request: Request = None, *args):
                if "/api" in request.url.path:
                    # programmatic api access, return json
                    return JSONResponse(
                        {
                            "detail": self.unauthorized_status_message,
                            "status_code": 403,
                        },
                        status_code=403,
                    )
                return app.templates.TemplateResponse(
                    "login.html.j2",
                    {
                        "request": request,
                        "api_key_name": self._api_key_name,
                        "status_code": 403,
                        "detail": self.unauthorized_status_message,
                    },
                )



import asyncio
from datetime import timedelta
from logging import getLogger
from typing import Any, Dict, List, Optional, Set, Tuple, TypeVar, get_args, get_origin
from uuid import uuid4

import csp
import janus
from csp import ts
from csp.impl.genericpushadapter import GenericPushAdapter
from pydantic import Field, PrivateAttr, TypeAdapter
from starlette.websockets import WebSocket, WebSocketDisconnect
from uvicorn.protocols.utils import ClientDisconnected

from csp_gateway.server import ChannelSelection, GatewayChannels, GatewayModule

# separate to avoid circular
from csp_gateway.server.web import (
    GatewayWebApp,
    get_default_responses,
    prepare_response,
)
from csp_gateway.utils import is_gateway_struct_like

log = getLogger(__name__)

T = TypeVar("T")
U = TypeVar("U")


class MountWebSocketRoutes(GatewayModule):
    requires: Optional[ChannelSelection] = []
    selection: ChannelSelection = Field(default_factory=ChannelSelection)

    readonly: bool = False
    prefix: str = "/stream"
    debug: bool = False
    ping_time_s: int = 1

    _channels: Any = PrivateAttr(None)
    _supported_channels: Dict[Tuple[str, Any], Any] = PrivateAttr(default_factory=dict)
    _dict_basket_keys_and_type_adapter: Dict[str, Tuple[Any, List[Any]]] = PrivateAttr(
        default_factory=dict
    )  # Maps dict basket channels to the type adapter for their keys and keys themselves
    _connect_events: Dict[Tuple[str, Any], GenericPushAdapter] = PrivateAttr(default_factory=dict)
    _disconnect_events: Dict[Tuple[str, Any], GenericPushAdapter] = PrivateAttr(default_factory=dict)
    _subscriptions: Dict[str, Set[Tuple[str, Any]]] = PrivateAttr(default_factory=dict)
    _queue: Optional[janus.Queue] = PrivateAttr(None)
    _task: Optional[asyncio.Task] = PrivateAttr(None)

    def connect(self, channels: GatewayChannels) -> None:
        self._channels = channels

        for field in self.selection.select_from(channels):
            maybe_edge = channels.get_channel(field)

            # Process both regular channels and dict baskets
            edges_and_keys = []
            if isinstance(maybe_edge, dict):
                # For dict baskets, collect each key and edge
                type_adapter = None
                keys_for_channel = []
                for key, edge in maybe_edge.items():
                    if type_adapter is None:
                        type_adapter = TypeAdapter(key.__class__)
                    edges_and_keys.append((key, edge))
                    keys_for_channel.append(key)
                self._dict_basket_keys_and_type_adapter[field] = (type_adapter, keys_for_channel)
            else:
                # Regular channel
                edges_and_keys = [(None, maybe_edge)]

            is_list_model, ts_type = None, None
            for key, edge in edges_and_keys:
                # Only assign is_list_model and
                # ts_type once
                if is_list_model is None:
                    # extract timeseries
                    ts_type = edge.tstype.typ

                    # check if its a list model
                    if get_origin(ts_type) is list:
                        is_list_model = True
                        ts_type = get_args(ts_type)[0]
                    else:
                        is_list_model = False

                channel_and_key = (field, key)
                name = field if key is None else f"{field}_{key}"

                # add to tracking
                self._supported_channels[channel_and_key] = (is_list_model, ts_type)

                # add generic push for connects
                self._connect_events[channel_and_key] = GenericPushAdapter(Tuple[str, WebSocket], name=f"connect_{name}")

                # add generic push for disconnects
                self._disconnect_events[channel_and_key] = GenericPushAdapter(Tuple[str, WebSocket], name=f"disconnect_{name}")

                # run csp node
                self.handle_websocket_connection(
                    edge,
                    connect=self._connect_events[channel_and_key].out(),
                    disconnect=self._disconnect_events[channel_and_key].out(),
                    channel_and_key=channel_and_key,
                    is_list_model=is_list_model,
                )

        heartbeat_key = ("heartbeat", None)
        self._supported_channels[heartbeat_key] = (False, None)
        self._connect_events[heartbeat_key] = GenericPushAdapter(Tuple[str, WebSocket], name="connect_heartbeat")
        self._disconnect_events[heartbeat_key] = GenericPushAdapter(Tuple[str, WebSocket], name="disconnect_heartbeat")
        self.handle_heartbeat_connection(
            connect=self._connect_events[heartbeat_key].out(),
            disconnect=self._disconnect_events[heartbeat_key].out(),
        )

    def start_app(self):
        """Start the app, needs to be called with event loop from main thread running."""
        if self._queue is None:
            self._queue = janus.Queue()
            self._task = asyncio.create_task(self.process_queue())

    async def process_queue(self):
        while True:
            try:
                uuid, websocket, message = await asyncio.wait_for(self._queue.async_q.get(), timeout=3.0)
                if uuid in self._subscriptions:
                    try:
                        if self.debug:
                            await websocket.send_text(message)
                        else:
                            await self.try_send_text(websocket, message)
                    except (ClientDisconnected, WebSocketDisconnect, asyncio.CancelledError) as e:
                        log.debug(f"Writing to client with {uuid = } disconnected with error {e = }")
                        self.disconnect_websocket(websocket, uuid)
            except TimeoutError:
                log.debug("No data to submit via websockets, retrying...")
            except Exception:
                log.exception("Something bad happened during queue processing")

    @csp.node
    def handle_heartbeat_connection(self, connect: ts["T"], disconnect: ts["T"]):
        with csp.alarms():
            a_send_ping: ts[bool] = csp.alarm(bool)
        with csp.state():
            s_connections: Dict[str, WebSocket] = {}

        with csp.start():
            csp.schedule_alarm(a_send_ping, timedelta(seconds=self.ping_time_s), True)

        if csp.ticked(connect):
            s_connections[connect[0]] = connect[1]

        if csp.ticked(disconnect):
            s_connections.pop(disconnect[0], None)

        if csp.ticked(a_send_ping):
            csp.schedule_alarm(a_send_ping, timedelta(seconds=self.ping_time_s), True)
            response = '{"channel": "heartbeat", "data": "PING"}'
            for uuid, websocket in s_connections.items():
                self._queue.sync_q.put((uuid, websocket, response))

    @csp.node
    def handle_websocket_connection(
        self,
        data: ts["T"],
        connect: ts["U"],
        disconnect: ts["U"],
        channel_and_key: Tuple[str, Any],  # channel name to possible dict basket key
        is_list_model: bool = False,
    ):
        with csp.state():
            s_connections: Dict[str, WebSocket] = {}
        with csp.start():
            # make the input passive unless we have active subscriptions
            csp.make_passive(data)

        if csp.ticked(connect):
            if len(s_connections) == 0:
                # make the data input active if this is the first connection
                csp.make_active(data)
            # add
            s_connections[connect[0]] = connect[1]

        if csp.ticked(disconnect):
            s_connections.pop(disconnect[0], None)

            if len(s_connections) == 0:
                # make the data input passive if that was the last connection
                csp.make_passive(data)

        if csp.ticked(data):
            channel, maybe_db_key = channel_and_key
            if maybe_db_key is None:
                key_json_fragment = ""
            else:
                key_json_fragment = f'"key": "{maybe_db_key.name if hasattr(maybe_db_key, "name") else maybe_db_key}",'
            # convert to pydantic, then convert to json, then send to all clients
            response = (
                f'{{"channel":"{channel}",'
                f"{key_json_fragment}"
                f'"data":{prepare_response(data, is_list_model=is_list_model, is_dict_basket=False, wrap_in_response=False)}}}'
            )
            for uuid, websocket in s_connections.items():
                self._queue.sync_q.put((uuid, websocket, response))

    async def try_send_text(self, websocket: WebSocket, resp: str):
        try:
            await websocket.send_text(resp)
        except (RuntimeError, AssertionError):
            log.exception(f"Error during websocket data send to CSP for {resp = }")

    def _handle_subscription(self, action: str, channel_and_key: Any, uuid: str, websocket: WebSocket):
        """Helper method to handle subscribe/unsubscribe for a specific channel+key"""
        if action == "subscribe" and channel_and_key not in self._subscriptions[uuid]:
            # add to subscriptions list

            self._subscriptions[uuid].add(channel_and_key)

            # push tick to csp
            self._connect_events[channel_and_key].push_tick((uuid, websocket))

        elif action == "unsubscribe" and channel_and_key in self._subscriptions[uuid]:
            # remove from subscriptions list
            self._subscriptions[uuid].discard(channel_and_key)

            # push tick to csp
            self._disconnect_events[channel_and_key].push_tick((uuid, websocket))

    def handle_message(self, message: dict, uuid: str, websocket: WebSocket):
        action = message.get("action", None)

        # validate
        if not action or action not in ("subscribe", "unsubscribe", "send"):
            # ignore
            log.info(f"unsupported action: {action}")
            return
        # get channel
        channel = message.get("channel", None)

        if not channel:
            # ignore
            log.info("no channel specified")
            return

        dict_basket_key_type_adapter, dict_basket_keys = self._dict_basket_keys_and_type_adapter.get(channel, (None, []))

        if dict_basket_keys and "key" not in message:
            if action in ("subscribe", "unsubscribe"):
                # Apply the action to all keys in the dict basket
                for key in dict_basket_keys:
                    self._handle_subscription(action, (channel, key), uuid, websocket)
                return
            elif action == "send":
                # For sending to a dict basket, we require a key
                log.info(f"sending data to dict basket {channel} requires a key")
                return

        raw_key = message.get("key", None)
        if raw_key and not dict_basket_key_type_adapter:
            log.info(f"Provided key {raw_key} for non-dict-basket channel {channel}, ignoring")
            return

        if dict_basket_key_type_adapter:
            key = dict_basket_key_type_adapter.validate_python(raw_key)
        else:
            key = raw_key

        # Create channel key tuple
        channel_and_key = (channel, key)

        if channel_and_key not in self._supported_channels:
            # ignore
            if key is None:
                log.info(f"unsupported channel: {channel}")
            else:
                log.info(f"unsupported channel: {channel} (dict-basket) with key: {key}")
            return

        if action == "send" and not self.readonly:
            # get the channel type from before
            is_list, channel_type = self._supported_channels.get(channel_and_key)

            # do some error checking
            if not isinstance(channel_type, type):
                log.info(f"unsupported type: {channel_type}")
                return

            if "data" not in message:
                log.info(f"unsupported data: {message}")
                return

            if is_gateway_struct_like(channel_type):
                # parse out the object and send
                msg_data = message.get("data")

                if not msg_data:
                    log.error("Error - received no data in websocket message")
                    return

                if not isinstance(msg_data, list):
                    msg_data = [msg_data]

                # do conversion in its own little try/except to avoid ambiguity
                try:
                    type_adapter = channel_type.type_adapter()
                    datum = [type_adapter.validate_python(obj) for obj in msg_data]
                except Exception:
                    log.error("Error during websocket data conversion", exc_info=True)
                    return

                try:
                    if is_list:
                        self._channels.send(channel, datum, key)
                    else:
                        for obj in datum:
                            self._channels.send(channel, obj, key)
                except Exception:
                    log.error("Error during websocket data send to CSP", exc_info=True)
                    return

            else:
                log.info(f"unsupported data: {message}")
        self._handle_subscription(action, channel_and_key, uuid, websocket)

    def disconnect_websocket(self, websocket: WebSocket, uuid: str):
        for subscription in self._subscriptions.get(uuid, set()):
            self._disconnect_events[subscription].push_tick((uuid, websocket))

        # remove from dict
        self._subscriptions.pop(uuid, None)

    async def websocket_handler(self, websocket: WebSocket):
        self.start_app()
        try:
            # connect to websocket
            await websocket.accept()
            uuid = str(uuid4())
            self._subscriptions[uuid] = set()
            log.debug(f"Client with {uuid = } connected")
            while True:
                try:
                    # get subscription message from client
                    message = await websocket.receive_json()
                    self.handle_message(message, uuid, websocket)

                except (ValueError, KeyError):
                    log.info(f"Malformed websocket message or error during data parsing: {message}", exc_info=True)
                    # ignore malformed
                    continue

        except (ClientDisconnected, WebSocketDisconnect, asyncio.CancelledError) as e:
            # push disconnect tick
            log.debug(f"Reading from client with {uuid = } disconnected with error {e = }")
        except Exception:
            log.exception("Error during websocket connection")

        finally:
            self.disconnect_websocket(websocket, uuid)

    def rest(self, app: GatewayWebApp) -> None:
        api_router = app.get_router("api")

        # add websocket handler
        api_router.add_api_websocket_route(path=self.prefix, endpoint=self.websocket_handler, name="streaming")

        # add dummy route for openapi
        @api_router.get(
            f"{self.prefix}",
            responses=get_default_responses(),
            response_model=List[str],
            tags=["Streaming"],
        )
        async def stream() -> List[str]:
            """
            This endpoint returns all of the available mounted websocket streams.

            The websocket stream is available on the same route.
            Data is bidirectional, any channel available as a `send` or `last` route is also available under the websocket connection.

            To subscribe to channels, send a JSON message of the form:
            ```
            {
                "action": "subscribe",
                "channel": "<channel name>",
                "key": "<optional dict basket key>"
            }
            ```

            For dict basket channels, omitting the "key" will subscribe to all available keys in that basket.

            To unsubscribe to channels, send a JSON message of the form:
            ```
            {
                "action": "unsubscribe",
                "channel": "<channel name>",
                "key": "<optional dict basket key>"
            }
            ```

            For dict basket channels, omitting the "key" will unsubscribe from all keys in that basket.

            Data will be sent across the websocket for all subscribed channels. It has the form:

            ```
            {
                "channel": "<channel name>",
                "key": "<only included for dict basket channels>",
                "data": <the same data that would be transmitted over e.g. the `last` endpoint>
            }
            ```
            """
            # Return a list of available channels
            res = []
            for channel, key in self._supported_channels:
                if key is None:
                    res.append(channel)
                else:
                    key_str = key.name if hasattr(key, "name") else key
                    res.append(f"{channel}/{key_str}")
            return res
